/**
 *  ===  WARNING  ==========================================<br/>
 *  This package is scheduled to be refactored and integrated<br/>
 *  in another package.<br/>
 *  Please keep changes minimal (bug fixes only) until then.<br/>
 *  ========================================================
 */
package com.adobe.epubcheck.ctc;