package com.adobe.epubcheck.vocab;

public final class MagazineNavigationVocab
{
  public static final String PREFIX = "msv";
  public static final String URI = "http://www.idpf.org/epub/vocab/structure/magazine/#";
  public static final Vocab VOCAB = new UncheckedVocab(URI, PREFIX);


  private MagazineNavigationVocab()
  {
  }

}
